# -*- coding: utf-8 -*-

# Calling Blender is currently a messy business because there's no easy
# way to install it as a python module. Instead we need to install Blender
# separately and have it execute scripts from the command-line. To do this,
# we need to check and see if Blender is installed, and if not, warn the user.
# This is different for each operating system, and each is outlined here:
#   https://docs.blender.org/manual/en/latest/advanced/command_line/launch/index.html

import os
import platform
import subprocess
from pathlib import Path

import yaml

# NOTE: You should only need to use the get_blender_command() function below.
# All other functions are called within it.
#
# The exception to this is if you installed a new version of Blender when you
# already had a older version configured. In this case, just run the
# find_blender_installation once and you're good to go!


def get_blender_command():

    # First we look in the simmate configuration directory and check if it
    # has been set there.
    #   [home_directory] ~/simmate/blender.yaml
    blender_filename = os.path.join(Path.home(), "simmate", "blender.yaml")
    if os.path.exists(blender_filename):
        with open(blender_filename) as file:
            blender_command = yaml.full_load(file)["COMMAND"]

        # BUG: for windows, we need to add quotes around the command because
        # it's actually a path. I couldn't figure out how to store these quotes
        # directly in the yaml file...
        if platform.system() == "Windows":
            blender_command = f'"{blender_command}"'

    # If not, we then try to find blender using the find_blender_installation
    # function defined below.
    else:
        blender_command = find_blender_installation()

    test_blender_command(blender_command)

    # we should now have the blender command and can return it for use elsewhere
    return blender_command


def find_blender_installation():
    """
    Finds the full path to the Blender installation so that we can call blender
    from the command-line. This also adds it's location to the simmate
    configuration folder so that we don't need to search for it every time.
    """

    # The operating system (Windows, OSX, or Linux) will give us the best guess
    # as to where to access the blender command. The results are as you'd expect
    # except for Macs, which return "Darwin".
    #   Linux: Linux
    #   Mac: Darwin
    #   Windows: Windows
    operating_system = platform.system()

    # Linux includes all distros where we primarily test with Ubuntu.
    if operating_system == "Linux":

        # On Linux, we actually have the easiest setup because the blender
        # command should be directly accessible -- regardless of where blender
        # was installed.
        blender_command = "blender"

    # Windows -- we assume Windows 10 for now.
    elif operating_system == "Windows":

        # On Windows, Blender is typically installed in the following directory:
        #   C:\Program Files\Blender Foundation\Blender 2.93
        expected_folder = "C:\\Program Files\\Blender Foundation\\"
        # in the future, I could have a list of directories to check here and
        # then iterate through them below.

        # We check the folder, and if it exists we grab the highest Blender version listed.
        if os.path.exists(expected_folder):

            # grab all the folders and find the highest blender version, which
            # be the last one when sorted alphabetically
            blender_version = sorted(os.listdir(expected_folder))[-1]

            # To call blender in python, we write the full path to the blender
            # executable. For example, the command looks like this:
            #   "C:\\Program Files\\Blender Foundation\\Blender 2.93\\blender.exe" --help
            full_path_to_blender = os.path.join(
                expected_folder, blender_version, "blender.exe"
            )
            # Because there are spaces in the file path, it is important we
            # wrap this command in quotes. Note we also replace the backslashes
            # in the file path with forward slashes.
            blender_command = f'"{full_path_to_blender}"'.replace("\\", "/")

        # if this path doesn't exist, we need to tell the user to install
        # Blender before than can use any of this functionality.
        else:
            raise BlenderNotInstalledError

    # Mac -- I don't own a Mac so I haven't implemented this yet.
    elif operating_system == "Darwin":
        raise NotImplementedError(
            "Simmate has not yet added support for Blender on Macs."
            " We can add this right away if you'd like! Just post a request"
            " on our forum."
        )

    # now confirm we can call the blender command
    test_blender_command(blender_command)

    # We should now have our blender_command set from above. So that we don't need
    # to search for the Blender installation every time we call a visualization
    # function, we want to save the path to a configuration file in /simmate

    # We store the command inside the config directory which is located at...
    #   [home_directory] ~/simmate/blender.yaml
    blender_filename = os.path.join(Path.home(), "simmate", "blender.yaml")
    with open(blender_filename, "w") as file:
        file.write(f"COMMAND: {blender_command}")

    return blender_command


def test_blender_command(blender_command):
    # To confirm we have blender configured properly, we often need to test it
    # out. We do this by calling the "blender --help" command
    command_to_test = f"{blender_command} --version"
    result = subprocess.run(
        command_to_test,
        shell=True,  # to access commands in the path
        capture_output=True,  # capture any ouput + error logs
    )

    # see if the command works, which means the return code is 0
    if result.returncode != 0:
        raise BlenderNotInstalledError


class BlenderNotInstalledError(Exception):
    def __init__(self):
        default_message = (
            "Make sure you have Blender installed before trying to access "
            "any of Simmate's 3D visualization features. If you have already "
            "installed Blender and are seeing this message, then please contact "
            "our team so we can help resolve the issue. You can download Blender "
            "from their website: https://www.blender.org/download/"
        )
        super().__init__(default_message)



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

import os

from . import settings  # configures settings but has no functionality
from .client import get_dask_client
from .submit import batch_submit



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import PositiveEnergy


def test_positive_energy(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="positive_energy",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    oszicar_filename = os.path.join(tmpdir, "OSZICAR")

    # init class with default settings
    error_handler = PositiveEnergy()

    # Confirm an error IS found using test files
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from Fast to Normal"
    assert Incar.from_file(incar_filename)["ALGO"] == "Normal"

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "halved POTIM from 0.5 to 0.25"
    assert Incar.from_file(incar_filename)["POTIM"] == 0.25

    # Confirm an error IS NOT found when the first ionic step hasn't completed yet
    # To check this, rewrite the OSZICAR file to only have the first line
    with open(oszicar_filename, "r") as file:
        contents = file.readlines()
    with open(oszicar_filename, "w") as file:
        file.writelines(contents[0])
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS NOT found when there is no OSZICAR
    os.remove(oszicar_filename)
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Rhosyg(ErrorHandler):

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["RHOSYG"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # We make a fix only if SYMPREC is not the default value
        if incar.get("SYMPREC", 1e-4) == 1e-4:
            incar["ISYM"] = 0
            correction = "switched ISYM to 0"
        else:
            correction = ""

        incar["SYMPREC"] = 1e-4
        correction += "switched SYMPREC back to default (1e-4)"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from pymatgen.core import Species

from simmate.toolkit import Structure
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.tasks.static_energy import MatprojStaticEnergy


class MatprojNMRElectricFieldGradiant(MatprojStaticEnergy):
    """
    This task is a reimplementation of pymatgen's
    [MPNMRSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPNonSCFSet)
    with mode="efg" (Electric Field Gradient).
    """

    incar = MatprojStaticEnergy.incar.copy()
    incar.update(
        dict(
            ALGO="FAST",
            EDIFF=-1.0e-10,
            ISYM=0,
            LCHARG=False,
            LEFG=True,
            QUAD_EFG__smart_quad_efg=True,
            NELMIN=10,
            PREC="ACCURATE",
            SIGMA=0.01,
        )
    )
    incar.pop("EDIFF__per_atom")


def keyword_modifier_smart_quad_efg(structure: Structure, quad_efg_config):
    """
    When running NMR (nuclear magnetic resonance) calculations that evaluate
    the electric field gradient, we need to set quadrapole moments for
    each species in our structure.

    NOTE: this modifier may be converted to a workflow parameter in the
    future.
    """
    # This code is copy/pasted from pymatgen (with minor edits)
    # See..
    # https://github.com/materialsproject/pymatgen/blob/1b6d1d2212dcf3a559cb2c489dd25e9754f9f788/pymatgen/io/vasp/sets.py#L1940-L1942

    # TODO: maybe use quad_efg_config to set this for different isotopes.
    # For now, I just use the VASP default
    isotopes = []

    isotopes = {ist.split("-")[0]: ist for ist in isotopes}
    quad_efg = [
        Species(element.symbol).get_nmr_quadrupole_moment(
            isotopes.get(element.symbol, None)
        )
        for element in structure.composition.elements
    ]

    # there is a unit (mbarn) attached to each so we clear these
    quad_efg = [float(q) for q in quad_efg]

    return quad_efg


Incar.add_keyword_modifier(keyword_modifier_smart_quad_efg)



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation import (
    MatVirtualLabCINEBEndpointRelaxation,
)


class NEBEndpointStaticEnergy(MatVirtualLabCINEBEndpointRelaxation):
    """
    Runs a VASP energy calculation using MIT Project settings, where some
    settings are adjusted to accomodate large supercells with defects.

    This is identical to relaxation/neb_endpoint, but just a single ionic step.

    You typically shouldn't use this workflow directly, but instead use the
    higher-level NEB workflows (e.g. diffusion/neb_all_paths or
    diffusion/neb_from_endpoints), which call this workflow for you.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    incar = MatVirtualLabCINEBEndpointRelaxation.incar.copy()
    incar.update(
        dict(
            IBRION=-1,  # (optional) locks everything between ionic steps
            NSW=0,  # this is the main static energy setting
        )
    )



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import Workflow
from simmate.workflow_engine.common_tasks import load_input_and_register

from simmate.calculators.vasp.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj as static_workflow,
)
from simmate.calculators.vasp.workflows.electronic_structure.matproj_density_of_states import (
    ElectronicStructure__Vasp__MatprojDensityOfStates as dos_workflow,
)
from simmate.calculators.vasp.workflows.electronic_structure.matproj_band_structure import (
    ElectronicStructure__Vasp__MatprojBandStructure as bandstruct_workflow,
)


class ElectronicStructure__Vasp__MatprojFull(Workflow):
    """
    Runs a static energy calculation followed by non-SCF calculations for
    band structure and density of states.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - static-energy.vasp.matproj
        - band-structure.vasp.matproj
        - density-of-states.vasp.matproj

    Note, these calculations are done using PBE, which is known to underestimate
    band gaps. For higher quality electronic structure calculations, you should
    use HSE instead.
    """

    description_doc_short = "runs DOS and BS at Materials Project settings"

    # table not implemented yet. This is a placeholder
    database_table = bandstruct_workflow.database_table

    @classmethod
    def run_config(
        cls,
        structure,
        command: str = None,
        source: str = None,
        directory: str = None,
        copy_previous_directory: str = False,
    ):

        parameters_cleaned = load_input_and_register(
            structure=structure,
            command=command,
            source=source,
            directory=directory,
            copy_previous_directory=copy_previous_directory,
            register_run=False,
        )

        static_result = static_workflow.run(
            structure=parameters_cleaned["structure"],
            command=parameters_cleaned.get("command"),
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + static_workflow.name_full,
            source=parameters_cleaned["source"],
            # For band-structures, unit cells should be in the standardized format
            pre_standardize_structure=True,
        ).result()  # block until complete

        dos_state = dos_workflow.run(
            structure={
                "database_table": static_workflow.database_table.__name__,
                "directory": static_result["directory"],
            },
            command=parameters_cleaned.get("command"),
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + dos_workflow.name_full,
            copy_previous_directory=True,
        )

        bs_state = bandstruct_workflow.run(
            structure={
                "database_table": static_workflow.database_table.__name__,
                "directory": static_result["directory"],
            },
            command=parameters_cleaned.get("command"),
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + bandstruct_workflow.name_full,
            copy_previous_directory=True,
        )

        # block until complete
        dos_state.result()
        bs_state.result()

        # TODO:
        # workup dos_state.result() and bs_state.result() into single figure



"""
This module provides conversion between the Simmate ToolkitStructure object and 
pymatgen Structure objects. Because our toolkit class inherits from pymatgen,
this feature should rarely (if ever) be used.

``` python
from simmate.toolkit import Structure
from simmate.file_converters.structure.pymatgen import PyMatGenAdapter


initial_structure = Structure.from_file("example.cif")

# convert to PyMatgen
new_pmg_object = PyMatGenAdapter.get_pymatgen(structure)

# convert back to PyMatgen/Simmate
new_toolkit_object = PyMatGenAdapter.get_toolkit(new_pmg_object)
```
"""

from simmate.toolkit import Structure as ToolkitStructure
from pymatgen.core.structure import Structure as PyMatGenStructure


class PyMatGenAdapter:
    """
    Adaptor for conversion between the Simmate ToolkitStructure object and
    pymatgen Structure objects.
    """

    @staticmethod
    def get_pymatgen(structure: ToolkitStructure) -> PyMatGenStructure:
        """
        Returns PyMatGen Structure object from Simmate Structure object.
        """
        data = structure.as_dict()
        return PyMatGenStructure.from_dict(data)

    @staticmethod
    def get_toolkit(structure: PyMatGenStructure) -> ToolkitStructure:
        """
        Returns PyMatGen Structure object from Simmate Structure object.
        """
        data = structure.as_dict()
        return ToolkitStructure.from_dict(data)



# -*- coding: utf-8 -*-

import os


def set_default_executor(dask_scheduler_address):
    """
    If you want to start a Prefect Agent that uses a specific Dask Cluster for
    all workflow runs, you can run this function before starting your Prefect
    Agent.

    What it does is set two enviornment variables that tell Prefect to default
    all workflows to using a default executor. So this saves us from having to
    repeatedly use this line below when setting ups workflows...
        from prefect.executors import DaskExecutor
        workflow.executor = DaskExecutor(address="tcp://152.2.172.72:8786")

    After you run this command, you can start your Prefect Agent as usual...
        from prefect.agent.local import LocalAgent
        agent = LocalAgent(name="ExampleAgent")
        agent.start()
    """

    # All workflows should be pointed to the Dask cluster as the default Executor.
    # We can grab the Dask scheduler's address using the cluster object from above.
    # For the master node, this address is "tcp://152.2.172.72:8786"
    os.environ.setdefault(
        "PREFECT__ENGINE__EXECUTOR__DEFAULT_CLASS", "prefect.executors.DaskExecutor"
    )
    os.environ.setdefault(
        "PREFECT__ENGINE__EXECUTOR__DASK__ADDRESS", dask_scheduler_address
    )



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import Brmix


def test_brmix(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="brmix",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    chgcar_filename = os.path.join(tmpdir, "CHGCAR")
    wavecar_filename = os.path.join(tmpdir, "WAVECAR")
    outcar_filename = os.path.join(tmpdir, "OUTCAR")
    errorcount_filename = os.path.join(tmpdir, "simmate_error_counts.json")

    # init class with default settings
    error_handler = Brmix()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasp.no_error"
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmpdir) == True

    # Handle error when there is a valid OUTCAR
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ISTART to 1"
    incar = Incar.from_file(incar_filename)
    assert incar["ISTART"] == 1

    # Remove the error counts and valid OUTCAR and restart
    os.remove(errorcount_filename)
    os.remove(outcar_filename)

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched IMIX to 1"
    incar = Incar.from_file(incar_filename)
    assert incar["IMIX"] == 1

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "removed any IMIX tag and switched KGAMMA to False"
    incar = Incar.from_file(incar_filename)
    assert incar.get("IMIX", None) == None
    assert incar["KGAMMA"] == False

    # Make third attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "removed any IMIX tag and switched KGAMMA to True"
    incar = Incar.from_file(incar_filename)
    assert incar.get("IMIX", None) == None
    assert incar["KGAMMA"] == True

    # Make final attempt at fixing the error
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ISYM to 0 and KGAMMA to True and deleted CHGCAR and WAVECAR"
    incar = Incar.from_file(incar_filename)
    assert incar["ISYM"] == 0
    assert incar["KGAMMA"] == True
    assert not os.path.exists(chgcar_filename)
    assert not os.path.exists(wavecar_filename)

    # Confirm an error IS NOT found when NELECT is in the INCAR
    with open(incar_filename, "w") as file:
        file.write("NELECT = 123")
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

import platform

import pytest

from simmate.visualization.structure.blender.configuration import (
    get_blender_command,
    BlenderNotInstalledError,
)

# Requires blender install which is not possible in CI yet.
#
# from simmate.visualization.structure.blender.base import (
#     make_blender_structure,
#     serialize_structure_sites,
# )


@pytest.mark.blender  # this test requires blender to NOT be installed
def test_get_blender_command():

    operating_system = platform.system()

    if operating_system == "Darwin":
        pytest.raises(NotImplementedError, get_blender_command)
    else:
        pytest.raises(BlenderNotInstalledError, get_blender_command)



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.band_structure import VaspBandStructure
from simmate.calculators.vasp.tasks.relaxation import MatprojHSERelaxation


class MatprojHSEBandStructure(VaspBandStructure, MatprojHSERelaxation):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet)
    with mode="line".

    Calculates the band structure using Materials Project HSE settings.
    """

    incar = MatprojHSERelaxation.incar.copy()
    incar.update(
        NSW=0,
        ISMEAR=0,
        SIGMA=0.05,
        ISYM=3,
        LCHARG=False,
        NELMIN=5,
    )



# -*- coding: utf-8 -*-

import os
from pathlib import Path
import shutil

import yaml

from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.vasp.outputs import Vasprun

from simmate.toolkit import Structure
from simmate.calculators.vasp.inputs import Incar, Poscar, Kpoints, Potcar
from simmate.workflow_engine import S3Task


def get_default_parallel_settings():
    """
    We load the user's default parallel settings from
        ~/simmate/vasp/INCAR_parallel_settings
    If this file doesn't exist, then we just use an empty dictionary.
    """
    settings_filename = os.path.join(
        Path.home(), "simmate", "vasp", "INCAR_parallel_settings"
    )
    if os.path.exists(settings_filename):
        return Incar.from_file(settings_filename)
    else:
        return {}


class VaspTask(S3Task):

    required_files = ["INCAR", "POTCAR", "POSCAR"]

    command: str = "vasp_std > vasp.out"
    """
    The command to call vasp, which is typically vasp_std. To ensure error
    handlers work properly, make sure your command has "> vasp.out" at the end.
    """
    # TODO: add support for grabbing a user-set default from their configuration
    # TODO: add auto check for vasp.out ending

    incar: dict = None
    """
    This sets the default vasp settings from a dictionary. This is the one thing
    you *must* set when subclassing VaspTask. An example is:
        
    ``` python
      incar = dict(NSW=0, PREC="Accurate", KSPACING=0.5)
    ```
    """

    incar_parallel_settings: dict = get_default_parallel_settings()
    """
    The parallel settings to add on to the base incar. These should not effect 
    the calculation result in any way (only how fast it completes), but they 
    are still selected based on the computer specs and what runs fastest on it.
    Therefore, these settings are loaded from ~/simmate/vasp/INCAR_parallel_settings
    by default and adding to this file should be the preferred method for updating
    these settings.
    """

    # TODO: add options for poscar formation
    # add_selective_dynamics=False
    # add_velocities=False
    # significant_figures=6 --> rounding issues? what's the best way to do this?

    kpoints = None
    """
    (experimental feature)
    The KptGrid or KptPath generator used to create the KPOINTS file. Note,
    this attribute is optional becuase VASP supports setting Kpts by adding
    KSPACING to the INCAR. If KSPACING is set in the INCAR, we ignore whatever 
    is set here.
    """
    # TODO - KptGrid is just a float for now, so there's no typing here.

    functional: str = None
    """
    This directs which Potcar files to grab. You would set this to a string
    of what you want, such as "PBE", "PBE_GW", or "LDA".
    """

    potcar_mappings: dict = None
    """
    This is an optional parameter to override Simmate's default selection of
    potentials based off of the functional chosen. The defaults are located
    in simmate.calculators.vasp.inputs.potcar_mappings. You can supply your
    own mapping dictionary or update the specific potentials you'd like. 
    For example:
    
    ``` python
      from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS
      element_mappings = PBE_ELEMENT_MAPPINGS.copy()  # don't forget to copy!
      element_mappings.update({"C": "C_h"})  # if you wish to update any
    ```
    
    or if you only use Carbon and don't care about other elements...
    
    ``` python
      element_mappings = {"C": "C_h"}
    ```
    
    Read more on this inside the Potcar class and be careful with updating!
    """

    confirm_convergence: bool = True
    """
    This flag controls whether or not we raise an error when the calculation 
    failed to converge. In somecases we still want results from calculations 
    that did NOT converge successfully.
    """
    # OPTIMIZE: What if I updated the ErrorHandler class to allow for "warnings"
    # instead of raising the error and applying the correction...? This functionality
    # could then be moved to the UnconvergedErrorHandler. I'd have a fix_error=True
    # attribute that is used in the .check() method. and If fix_error=False, I
    # simply print a warning & also add that warning to simmate_corrections.csv

    pre_sanitize_structure: bool = False
    """
    In some cases, we may want to sanitize the structure during our setup().
    This means converting to the LLL-reduced primitive cell. This simply does:
    ``` python
    structure_sanitzed = structure.copy(santize=True)
    ```
    """

    pre_standardize_structure: bool = False
    """
    In some cases, we may want to convert the structure to the standard primitive
    of a structure. For example, this is required when calculating band structures
    and ensuring we have a standardized high-symmetry path.
    """

    @classmethod
    def _get_clean_structure(
        cls,
        structure: Structure,
        pre_sanitize_structure: bool = None,
        pre_standardize_structure: bool = None,
        **kwargs,
    ) -> Structure:
        """
        Uses the class attributes for `pre_sanitize_structure` and
        `pre_standardize_structure`. If either of these are set to True, then
        the structure unitcell is converted using the proper methods.

        Note, this method is typically called within `setup` before any input
        files are written. You should never have to call it directly.
        """

        # See if these values were provided, or default to class attribute
        pre_sanitize_structure = pre_sanitize_structure or cls.pre_sanitize_structure
        pre_standardize_structure = (
            pre_standardize_structure or cls.pre_standardize_structure
        )

        # if both pre_standardize_structure and pre_sanitize_structure are set,
        # we raise an error. I may want to change this in the future though
        if pre_sanitize_structure and pre_standardize_structure:
            raise Exception(
                "For this VaspTask, only one of `pre_sanitize_structure` or "
                " `pre_standardize_structure` can be set to True, not both."
            )

        # If requested, we convert to the LLL-reduced unit cell, which aims to
        # be as cubic as possible.
        if pre_sanitize_structure:
            structure_cleaned = structure.copy(sanitize=True)
            return structure_cleaned

        # For band structures, we need to make sure the structure is in the
        # standardized primitive form.
        # We use the same SYMPREC from the INCAR, which is 1e-5 if not set.
        if pre_standardize_structure:
            sym_prec = cls.incar.get("SYMPREC", 1e-5) if cls.incar else 1e-5
            sym_finder = SpacegroupAnalyzer(structure, symprec=sym_prec)
            structure_cleaned = sym_finder.get_primitive_standard_structure(
                international_monoclinic=False,
            )
            # check for pymatgen bugs here
            check_for_standardization_bugs(structure, structure_cleaned)
            return structure_cleaned

        # if none of the flags above were used, then we just return the orignal
        # input structure.
        return structure

    @classmethod
    def setup(cls, directory: str, structure: Structure, **kwargs):

        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, os.path.join(directory, "POSCAR"))

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**cls.incar) + Incar(**cls.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            structure=structure_cleaned,
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            Kpoints.to_file(
                structure_cleaned,
                cls.kpoints,
                os.path.join(directory, "KPOINTS"),
            )

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure_cleaned.composition.elements,
            cls.functional,
            os.path.join(directory, "POTCAR"),
            cls.potcar_mappings,
        )

    @classmethod
    def setup_restart(cls, directory: str, **kwargs):
        """
        From a working directory of a past calculation, sets up for the calculation
        to be restarted. For relaxations/dynamics this involved just copying
        the poscar to the contcar.
        """

        # establish filenames
        poscar_filename = os.path.join(directory, "POSCAR")
        poscar_orig_filename = os.path.join(directory, "POSCAR_original")
        contcar_filename = os.path.join(directory, "CONTCAR")
        stopcar_filename = os.path.join(directory, "STOPCAR")

        # TODO:
        # make an archive of the directory before we start editting files
        # make_error_archive(directory)
        # add these changes to the simmate_corrections.csv

        # delete the stopcar if it exists
        if os.path.exists(stopcar_filename):
            os.remove(stopcar_filename)

        # copy poscar to a new file
        shutil.move(poscar_filename, poscar_orig_filename)

        # then CONTCAR over to the POSCAR
        shutil.move(contcar_filename, poscar_filename)

    @classmethod
    def workup(cls, directory: str):
        """
        This is the most basic VASP workup where I simply load the final structure,
        final energy, and (if requested) confirm convergence. I will likely make
        this a common function for this vasp module down the road.
        """

        # load the xml file and all of the vasprun data
        try:
            vasprun = Vasprun(
                filename=os.path.join(directory, "vasprun.xml"),
                exception_on_bad_xml=True,
            )
        except:
            print(
                "XML is malformed. This typically means there's an error with your"
                " calculation that wasn't caught by your ErrorHandlers. We try"
                " salvaging data here though."
            )
            vasprun = Vasprun(
                filename=os.path.join(directory, "vasprun.xml"),
                exception_on_bad_xml=False,
            )
            vasprun.final_structure = vasprun.structures[-1]
        # BUG: This try/except is 100% just for my really rough calculations
        # where I don't use any ErrorHandlers and still want the final structure
        # regarless of what when wrong. In the future, I should consider writing
        # a separate method for those that loads the CONTCAR and moves on.

        # write output files/plots for the user to quickly reference
        cls._write_output_summary(directory, vasprun)

        # confirm that the calculation converged (ionicly and electronically)
        if cls.confirm_convergence:
            assert vasprun.converged

        # return vasprun object
        return vasprun

    @staticmethod
    def _write_output_summary(directory: str, vasprun: Vasprun):
        """
        This prints a "simmate_summary.yaml" file with key output information.

        This method should not be called directly as it used within workup().
        """
        # OPTIMIZE: Ideally, I could take the vasprun object and run to_json,
        # but this output is extremely difficult to read.

        results = vasprun.as_dict()["output"]

        summary = {
            "structure_final": "The final structure is located in the CONTCAR file",
            "energy_final": float(results.get("final_energy", None)),
            "energy_final_per_atom": float(results.get("final_energy_per_atom", None)),
            "converged_electroinc": vasprun.converged_electronic,
            "converged_ionic": vasprun.converged_ionic,
            "fermi_energy": results.get("efermi", None),
            "valence_band_maximum": results.get("vbm", None),
            "conduction_band_minimum": results.get("vbm", None),
        }

        summary_filename = os.path.join(directory, "simmate_summary.yaml")
        with open(summary_filename, "w") as file:
            content = yaml.dump(summary)
            file.write(content)

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class. This is useful for printing out
        settings for users to inspect.
        """
        return {
            key: getattr(cls, key)
            for key in [
                "__module__",
                "pre_sanitize_structure",
                "confirm_convergence",
                "functional",
                "incar",
                "potcar_mappings",
            ]
        }


def check_for_standardization_bugs(
    structure_original: Structure,
    structure_new: Structure,
):

    # In pymatgen, they include this code with the standardization of their
    # structures because there were several bugs in the past and they want to
    # double-check themselves. I'm still using their code to standardize
    # my structures, so I should make this check too.

    vpa_old = structure_original.volume / structure_original.num_sites
    vpa_new = structure_new.volume / structure_new.num_sites

    if abs(vpa_old - vpa_new) / vpa_old > 0.02:
        raise ValueError(
            "Standardizing failed! Volume-per-atom changed... "
            f"old: {vpa_old}, new: {vpa_new}"
        )

    sm = StructureMatcher()
    if not sm.fit(structure_original, structure_new):
        raise ValueError("Standardizing failed! Old structure doesn't match new.")



# -*- coding: utf-8 -*-

"""
This file is for any custom Tasks you may need in your workflows. For example,
you may want to write a new VaspTask that has custom INCAR settings. These tasks
can be incorporated into our workflows (in `workflows.py`).

Below is an example of a simple VaspTask, which is used to run a single VASP
calculation.

For more information and advanced guides, be sure to read through our
[S3Task documentation](https://jacksund.github.io/simmate/simmate/workflow_engine/supervised_staged_shell_task.html)
as well as relevent subclasses like the
[VaspTask documentation](https://jacksund.github.io/simmate/simmate/calculators/vasp/tasks/base.html)
"""


from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)
from simmate.calculators.vasp.error_handlers import (
    Unconverged,
    NonConverging,
    Frozen,
    Walltime,
    # There are many more available too!
)


class ExampleRelaxation(VaspTask):

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria when available.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    # These are all input settings for this task.
    incar = dict(
        # These settings are the same for all structures regardless of composition.
        PREC="Normal",
        EDIFF=1e-5,
        ENCUT=450,  # !!! Should this be based on the element type?
        ISIF=3,
        NSW=100,
        IBRION=1,
        POTIM=0.02,
        LCHARG=False,
        LWAVE=False,
        KSPACING=0.4,
        # The type of smearing we use depends on if we have a metal, semiconductor,
        # or insulator. So we need to decide this using a keyword modifier.
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=1,
                SIGMA=0.06,
            ),
            "non-metal": dict(
                ISMEAR=0,
                SIGMA=0.05,
            ),
        },
        # This is a dummy example of using custom modifiers. This would set the
        # incar value of EXAMPLE=16 for structure with 2 sites (2*8=16). Note,
        # we registered "multiply_nsites" logic below.
        EXAMPLE__multiply_nsites=8,
    )

    # These are some default error handlers to use. Note the order matters here!
    # Only the first error handler triggered in this list will be used before
    # restarting the job
    error_handlers = [
        Unconverged(),
        NonConverging(),
        Frozen(),
        Walltime(),
    ]


# If you need to add advanced logic for one of your INCAR tags, you can register
# a keyword_modifier to the INCAR class like so:

# first define the logic of your modifier as a function
def keyword_modifier_multiply_nsites(structure, example_mod_input):
    # add your advanced logic to determine the keyword value.
    return structure.num_sites * example_mod_input


# then register modifier with the Incar class
Incar.add_keyword_modifier(keyword_modifier_multiply_nsites)



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class SymprecNoise(ErrorHandler):
    """
    Fixes determination of symmetry in problematic structures.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = [
        "determination of the symmetry of your systems shows a strong"
    ]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # make the fix
        if incar.get("ISYM", 2) > 0 and incar.get("SYMPREC", 1e-5) > 1e-6:
            incar["SYMPREC"] = 1e-6
            correction = "switched SYMPREC to 1e-6"
        else:
            incar["ISYM"] = 0
            correction = "switched ISYM to 0"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from django.urls import path
from simmate.website.third_parties import views


urlpatterns = [
    path(
        route="",
        view=views.providers_all,
        name="home",
    ),
    path(
        route="<provider_name>/",
        view=views.ProviderAPIViewSet.dynamic_list_view,
        name="provider",
    ),
    path(
        route="<provider_name>/<pk>/",
        view=views.ProviderAPIViewSet.dynamic_retrieve_view,
        name="entry-detail",
    ),
]



# -*- coding: utf-8 -*-

"""
This file connects to the Simmate database. Behind the scenes, this is really
setting up Django settings (which includes the database). Therefore, this 
file serves as a shortcut for the following:

``` python    
# the convenient shortcut
from simmate.database import connect

# the longer import that does the exact same thing
from simmate.configuration.django import setup
```
"""

# This import sets up Django so we can connect with the Simmate database
from simmate.configuration.django import setup



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.band_structure import (
    MatprojBandStructure as MatprojBandStructureTask,
)
from simmate.calculators.vasp.database.band_structure import (
    MatprojBandStructure as MatprojBandStructureResults,
)


class ElectronicStructure__Vasp__MatprojBandStructure(Workflow):
    s3task = MatprojBandStructureTask
    database_table = MatprojBandStructureResults
    description_doc_short = "uses Materials Project settings (PBE)"



# -*- coding: utf-8 -*-

import json
import sys
import argparse

import numpy

import bpy
import bmesh

# These are the RGB values that JMol uses to color atoms
JMOL_COLORING = {
    "H": [255, 255, 255],
    "He": [217, 255, 255],
    "Li": [204, 128, 255],
    "Be": [194, 255, 0],
    "B": [255, 181, 181],
    "C": [144, 144, 144],
    "N": [48, 80, 248],
    "O": [255, 13, 13],
    "F": [144, 224, 80],
    "Ne": [179, 227, 245],
    "Na": [171, 92, 242],
    "Mg": [138, 255, 0],
    "Al": [191, 166, 166],
    "Si": [240, 200, 160],
    "P": [255, 128, 0],
    "S": [255, 255, 48],
    "Cl": [31, 240, 31],
    "Ar": [128, 209, 227],
    "K": [143, 64, 212],
    "Ca": [61, 255, 0],
    "Sc": [230, 230, 230],
    "Ti": [191, 194, 199],
    "V": [166, 166, 171],
    "Cr": [138, 153, 199],
    "Mn": [156, 122, 199],
    "Fe": [224, 102, 51],
    "Co": [240, 144, 160],
    "Ni": [80, 208, 80],
    "Cu": [200, 128, 51],
    "Zn": [125, 128, 176],
    "Ga": [194, 143, 143],
    "Ge": [102, 143, 143],
    "As": [189, 128, 227],
    "Se": [255, 161, 0],
    "Br": [166, 41, 41],
    "Kr": [92, 184, 209],
    "Rb": [112, 46, 176],
    "Sr": [0, 255, 0],
    "Y": [148, 255, 255],
    "Zr": [148, 224, 224],
    "Nb": [115, 194, 201],
    "Mo": [84, 181, 181],
    "Tc": [59, 158, 158],
    "Ru": [36, 143, 143],
    "Rh": [10, 125, 140],
    "Pd": [0, 105, 133],
    "Ag": [192, 192, 192],
    "Cd": [255, 217, 143],
    "In": [166, 117, 115],
    "Sn": [102, 128, 128],
    "Sb": [158, 99, 181],
    "Te": [212, 122, 0],
    "I": [148, 0, 148],
    "Xe": [66, 158, 176],
    "Cs": [87, 23, 143],
    "Ba": [0, 201, 0],
    "La": [112, 212, 255],
    "Ce": [255, 255, 199],
    "Pr": [217, 255, 199],
    "Nd": [199, 255, 199],
    "Pm": [163, 255, 199],
    "Sm": [143, 255, 199],
    "Eu": [97, 255, 199],
    "Gd": [69, 255, 199],
    "Tb": [48, 255, 199],
    "Dy": [31, 255, 199],
    "Ho": [0, 255, 156],
    "Er": [0, 230, 117],
    "Tm": [0, 212, 82],
    "Yb": [0, 191, 56],
    "Lu": [0, 171, 36],
    "Hf": [77, 194, 255],
    "Ta": [77, 166, 255],
    "W": [33, 148, 214],
    "Re": [38, 125, 171],
    "Os": [38, 102, 150],
    "Ir": [23, 84, 135],
    "Pt": [208, 208, 224],
    "Au": [255, 209, 35],
    "Hg": [184, 184, 208],
    "Tl": [166, 84, 77],
    "Pb": [87, 89, 97],
    "Bi": [158, 79, 181],
    "Po": [171, 92, 0],
    "At": [117, 79, 69],
    "Rn": [66, 130, 150],
    "Fr": [66, 0, 102],
    "Ra": [0, 125, 0],
    "Ac": [112, 171, 250],
    "Th": [0, 186, 255],
    "Pa": [0, 161, 255],
    "U": [0, 143, 255],
    "Np": [0, 128, 255],
    "Pu": [0, 107, 255],
    "Am": [84, 92, 242],
    "Cm": [120, 92, 227],
    "Bk": [138, 79, 227],
    "Cf": [161, 54, 212],
    "Es": [179, 31, 212],
    "Fm": [179, 31, 186],
    "Md": [179, 13, 166],
    "No": [189, 13, 135],
    "Lr": [199, 0, 102],
    "Rf": [204, 0, 89],
    "Db": [209, 0, 79],
    "Sg": [217, 0, 69],
    "Bh": [224, 0, 56],
    "Hs": [230, 0, 46],
    "Mt": [235, 0, 38],
}

# Blender needs a 4th value for the opacity in addition to the RGB values given
# above. For all materials, we use 255 and append this to all of them here. Blender
# needs these values on a 0-1 scale instead of the 0-255. We address
# this below by dividing all values by 255
for key in JMOL_COLORING:
    color = JMOL_COLORING[key]
    color.append(255)
    JMOL_COLORING[key] = numpy.array(color) / 255


def make_structure_blend(lattice, sites_to_draw, filename):

    # convert variable from json str to original format
    lattice = json.loads(lattice)
    sites_to_draw = json.loads(sites_to_draw.replace("'", '"'))

    # import Verge3D settings
    # import addon_utils
    # addon_utils.enable(module_name="verge3d")

    # Clear existing objects.
    bpy.ops.wm.read_factory_settings(use_empty=True)

    # we grab the entire blender scene for reference as it let's us access
    # all objects later
    scene = bpy.context.scene

    # -------------------------------------------------------------------------

    # ADDING THE SITES

    # We start by drawing each of the sites -- which is just a colored sphere
    # at the proper coordinates
    for site in sites_to_draw:

        # first pull the base information out of the serialized tuple
        element_symbol, radius, cartesian_coords = site

        # we change the coordinates into a numpy array for functionality
        cartesian_coords = numpy.array(cartesian_coords)

        # Add a sphere for the site. Note we make the radius size only 0.75% its
        # true size in order to help with visualization.
        bpy.ops.mesh.primitive_ico_sphere_add(
            subdivisions=3,
            radius=radius * 0.75,
            location=cartesian_coords,
        )

        # Now we need to color and style the sphere.

        # grab the site color from our mappings above
        site_color = JMOL_COLORING[element_symbol]

        # first check if we have made this material already (i.e. an element of
        # this type has been made before). If so, we use that one.
        materials = bpy.data.materials
        if element_symbol in materials.keys():
            mat = materials[element_symbol]
        # otherwise we make a new material and name it after the element for
        # future reference.
        else:
            mat = bpy.data.materials.new(name=element_symbol)
            mat.diffuse_color = site_color
            mat.metallic = 1
            mat.specular_intensity = 0
            mat.roughness = 0.6
        # Now that we have the proper material create/selected, we can now
        # apply it to our sphere
        bpy.context.active_object.data.materials.append(mat)

    # We apply smooth shading to all the spheres and then deselect them before
    # moving on to the next step
    bpy.ops.object.select_all(action="SELECT")
    bpy.ops.object.shade_smooth()
    bpy.ops.object.select_all(action="DESELECT")

    # -------------------------------------------------------------------------

    # ADDING THE LATTICE

    # We make a lattice by creating a cube, deleting all of the faces, and then
    # manually placing each of its verticies to match the lattice size.
    bpy.ops.mesh.primitive_cube_add(size=1, enter_editmode=True)
    bpy.ops.mesh.delete(type="ONLY_FACE")
    bpy.ops.object.editmode_toggle()
    verts = bpy.context.object.data.vertices
    verts[0].co = (0, 0, 0)
    verts[1].co = lattice[2]
    verts[2].co = lattice[0]
    verts[3].co = numpy.add(lattice[0], lattice[2])
    verts[4].co = lattice[1]
    verts[5].co = numpy.add(lattice[1], lattice[2])
    verts[6].co = numpy.add(lattice[0], lattice[1])
    verts[7].co = numpy.sum(lattice, axis=0)

    # There's an issue where each lattice edge isn't a perfect line. To fix
    # this, we split the cube into separate lines and make sure each of those
    # are "full curves" which is really just a cylinder.

    # This is the easy want to do this with the UI but we get an error here...
    # bpy.ops.mesh.edge_split() # doesn't work because of context/poll check
    lattice = bpy.data.objects[0].data  # regular bpy object
    bm = bmesh.new()  # create new bmesh
    bm.from_mesh(lattice)  # fill bmesh with data from bpy object
    bmesh.ops.split_edges(bm, edges=bm.edges)  # spit the edges on the mesh
    bm.to_mesh(lattice)  # write the result data back to the initial bpy object

    # now fill each vector to a given size
    bpy.ops.object.convert(target="CURVE")
    bpy.context.object.data.fill_mode = "FULL"
    bpy.context.object.data.bevel_depth = 0.1
    bpy.context.object.data.bevel_resolution = 3
    bpy.ops.object.shade_smooth()

    # Now we create a black material to color the lattice with
    mat = bpy.data.materials.new(name="Lattice")
    mat.diffuse_color = (0, 0, 0, 1)
    mat.specular_intensity = 0
    bpy.context.active_object.data.materials.append(mat)

    # -------------------------------------------------------------------------

    # CENTERING ALL OBJECTS

    # When we created all the objects above, the center of the scene is (0,0,0)
    # for the cartesian coordinates, but it's better to have the viewpoint and
    # object rotation about the center of the lattice. Therefore, we grab the
    # center of the lattice, and use this location to translate all objects in
    # the scene such that this is the new center.
    bpy.ops.object.origin_set(type="ORIGIN_GEOMETRY", center="MEDIAN")
    lattice_center = bpy.data.objects["Cube"].location.copy()
    for obj in bpy.data.objects:
        obj.location = numpy.subtract(obj.location, lattice_center)

    # -------------------------------------------------------------------------

    # CONFIGURING THE REST OF THE SCENE

    # Camera
    cam_data = bpy.data.cameras.new(name="MyCam")
    cam_ob = bpy.data.objects.new(name="MyCam", object_data=cam_data)
    scene.collection.objects.link(cam_ob)  # instance the camera object in the scene
    scene.camera = cam_ob  # set the active camera
    cam_ob.rotation_euler = numpy.radians((70, 0, 93))
    cam_ob.location = (30, 2, 11)
    # cam_ob.data.type = 'ORTHO' # 'PERSP'

    # Sun
    light_data = bpy.data.lights.new("MyLight", "SUN")
    light_ob = bpy.data.objects.new(name="MyLight", object_data=light_data)
    scene.collection.objects.link(light_ob)

    # Set sun to move along with the camera. This is because we don't want
    # shadows changing in the viewport for crystal structures.
    light_ob.parent = cam_ob
    light_ob.location = (4, 50, 4)
    light_ob.rotation_euler = numpy.radians((60, 10, 150))

    # Background (aka the blender "World")
    world = bpy.data.worlds.new(name="MyWorld")
    world.color = (1, 1, 1)
    scene.world = world

    # -------------------------------------------------------------------------

    ## Center all objects at the origin  # fails as-is. consider centering camera to lattice
    # bpy.ops.object.select_all(action='SELECT')
    # bpy.ops.view3d.snap_selected_to_cursor(use_offset=True)

    ## scale the whole crystal structure
    # bpy.ops.object.select_all(action='SELECT')
    # bpy.ops.transform.resize(value=(1.29349, 1.29349, 1.29349))

    # update view to include all the changes we made above
    bpy.context.view_layer.update()

    # set verge3D settings
    # bpy.context.scene.v3d_export.use_shadows = False
    # bpy.context.scene.v3d_export.lzma_enabled = (
    #     True  # add compressed files (fails for some reason)
    # )
    # bpy.context.scene.v3d_export.aa_method = "MSAA8"
    # bpy.data.objects["MyCam"].data.v3d.orbit_min_distance = 15
    # bpy.data.objects["MyCam"].data.v3d.orbit_max_distance = 100
    #
    # export for Verge3D
    # bpy.ops.export_scene.v3d_gltf(filepath=save_path)

    # The format we save the file as depends on the ending of the filename
    if filename.endswith(".blend"):
        # save this to a blender file
        bpy.ops.wm.save_as_mainfile(filepath=filename)
    elif filename.endswith(".glb"):
        # export in the gltf 2.0 format (.glb file)
        bpy.ops.export_scene.gltf(filepath=filename)
    else:
        raise Exception("unknown format used")


def main():

    # get the arguments passed to blender after "--", all of which are ignored by
    # blender so scripts may receive their own arguments.
    arguments = sys.argv[sys.argv.index("--") + 1 :]

    # To pull out the arguments passed to the script, we need to tell the parser
    # what they will be in advance.
    parser = argparse.ArgumentParser()
    parser.add_argument("--lattice", dest="lattice")
    parser.add_argument("--sites", dest="sites")
    parser.add_argument("--save", dest="filename")

    # we can now pull out the arguments passed into the command
    parsed_arguments = parser.parse_args(arguments)

    # Run the function we defined above
    make_structure_blend(
        parsed_arguments.lattice,
        parsed_arguments.sites,
        parsed_arguments.filename,
    )


# This is boiler plate code that calls the main function when this script is
# ran with python directly.
if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from simmate.toolkit import Structure
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.tasks.dynamics import MITDynamics


class MatVirtualLabNPTDynamics(MITDynamics):
    """
    This task is a reimplementation of pymatgen's
    [MVLNPTMDSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLNPTMDSet).
    """

    confirm_convergence = False

    incar = MITDynamics.incar.copy()
    incar.update(
        dict(
            IALGO=48,
            ISIF=3,
            LANGEVIN_GAMMA__smart_langevin=True,
            LANGEVIN_GAMMA_L=1,
            MDALGO=3,
            PMASS=10,
            PSTRESS=0,
            SMASS=0,
            ENCUT=450,  # pymatgen sets to 1.5 * max of all potcars
        )
    )


def keyword_modifier_smart_langevin(
    structure: Structure,
    langevin_config: bool = True,  # not required
):
    """
    Expands LANGEVIN_GAMMA setting based on number of species present
    """
    return [10] * structure.ntypesp


Incar.add_keyword_modifier(keyword_modifier_smart_langevin)



# -*- coding: utf-8 -*-

"""
# The Supervised-Staged-Shell Task (aka "S3Task")

This class contains the core functionality to **supervise** a **staged** task
involving some **shell** command.

Let's breakdown what this means...

A *shell* command is a single call to some external program. For example,
VASP requires that we call the "vasp_std > vasp.out" command in order to run a
calculation. We consider calling external programs a *staged* task made
up of three steps:

- setup = writing any input files required for the program
- execution = actually calling the command and running our program
- workup = loading data from output files back into python

And for *supervising* the task, this means we monitor the program while the
execution stage is running. So once a program is started, Simmate can check
output files for common errors/issues. If an error is found, we stop the
program, fix the issue, and then restart it. Any fixes that were made are
written to "simmate_corrections.csv".


<!--
TODO: Make a simple diagram to visualize the overall process and add it here.
It will be similar to Custodian's, but we don't have a list of jobs here.
https://materialsproject.github.io/custodian/index.html#usage
The steps are...

- Write Input Files based on custom+defualt settings
- Run the calculation by calling the program
- Load ouput files
- check for errors
- [correct them, rerun]
- postprocess/analysis
-->

This entire process (the different stages and monitoring) is carried out
using the `run` method. You rarely use this class directly. Instead,
you typically use a subclass of it. As a user, you really just need to do
something like this:

``` python
from simmate.calculator.example.tasks import ExampleTask

my_task = ExampleTask()
my_result = my_task.run()
```

And that's it!

For experts, this class can be viewed as a combination of prefect's ShellTask,
a custodian Job, and Custodian monitoring. When subclassing this, we can absorb
functionality of `pymatgen.io.vasp.sets` too. By merging all of these together
into one class, we make things much easier for users and creating new Tasks.


# Building a custom S3task

This class is commonly used to make tasks for our calculator modules, so you
will likely want to subclass this. Here is a basic example of inheriting
and then running a task:

``` python

from simmate.workflow_engine import S3Task
from example.error_handlers import PossibleError1, PossibleError2


class ExampleTask(SSSTask):

    command = "echo example"  # just prints out "example"
    
    # settings for error handling
    max_corrections = 7
    error_handlers = [PossibleError1, PossibleError2]
    polling_timestep = 0.1
    monitor_freq = 10
    
    # custom attributes
    some_new_setting = 123

    @classmethod
    def setup(cls, directory, custom_parmeter, **kwargs):
        print("I'm setting things up!")
        print(f"My new setting value is {cls.some_new_setting}")
        print(f"My new parameter value is {custom_parmeter}")

    @staticmethod
    def workup(directory):
        print("I'm working things up!")


task = ExampleTask()
result = task.run()
```

There are a two important things to note here:

1. It's optional to write new `setup` or `workup` methods. But if you do...
    - Both `setup` and `workup` method should be either a staticmethod or classmethod
    - Custom `setup` methods require the `directory` and `**kwargs` input parameters.
    - Custom `workup` methods require the `directory` input paramter
2. It's optional to set/overwrite attributes. You can also add new ones too.

For a full (and advanced) example, of a subclass take a look at
`simmate.calculators.vasp.tasks.base.VaspTask` and the tasks that use it like
`simmate.calculators.vasp.tasks.relaxation.matproj`.
"""

import os
import platform
import time
import signal
import subprocess
import yaml
from typing import List, Tuple
from functools import cache

import pandas

from prefect.tasks import Task
from prefect.context import get_run_context, MissingContextError

from simmate.workflow_engine import ErrorHandler
from simmate.utilities import get_directory, make_archive, make_error_archive

# cleanup_on_fail=False, # TODO I should add a Prefect state_handler that can
# reset the working directory between task retries -- in some cases we may
# want to delete the entire directory.

# OPTIMIZE: I think this class would greatly benefit from asyncio so that we
# know exactly when a shelltask completes, rather than looping and checking every
# set timestep.
# https://docs.python.org/3/library/asyncio-subprocess.html#asyncio.create_subprocess_exec


class S3Task:
    """
    The base Supervised-Staged-Shell that many tasks inherit from. This class
    encapulates logic for running external programs (like VASP), fixing common
    errors during a calculation, and working up the results.
    """

    # I set this here so that I don't have to copy/paste the init method
    # every time I inherit from this class and want to update the default
    # command to use for the child class.
    command: str = None
    """
    The defualt shell command to use.
    """

    required_files = []
    """
    Before the command is executed, this list of files should be present
    in the working directory. This check will be made after `setup` is called
    and before `execute` is called. By default, no input files are required.
    """

    error_handlers: List[ErrorHandler] = []
    """
    A list of ErrorHandler objects to use in order of priority (that is, highest
    priority is first). If one handler is triggered, the correction will be 
    applied and none of the following handlers will be checked.
    """

    max_corrections: int = 10
    """
    The maximum number of times we can apply a correction and retry the shell
    command. The maximum number of times that corrections will be made (and
    shell command reran) before giving up on the calculation. Note, once this
    limit is exceeded, the error is stored without correcting or restarting
    the run.
    """

    monitor: bool = True
    """
    Whether to run monitor handlers while the command runs. False means
    wait until the job has completed.
    """

    polling_timestep: float = 1
    """
    If we are monitoring the job for errors while it runs, this is how often
    (in seconds) we should check the status of our job. Note this check is
    just whether the job is done or not. This is NOT how often we check for
    errors. See monitor_freq for that.
    """

    monitor_freq: int = 300
    """
    The frequency we should run check for errors with our monitors. This is
    based on the polling_timestep loops. For example, if we have a
    polling_timestep of 10 seconds and a monitor_freq of 2, then we would run
    the monitor checks every other loop -- or every 2x10 = 20 seconds. The
    default values of polling_timestep=10 and monitor_freq=30 indicate that
    we run monitoring functions every 5 minutes (10x30=300s=5min).
    """

    save_corrections_to_file: bool = True
    """
    Whether to write a log file of the corrections made. The default is True.
    """

    compress_output: bool = False
    """
    Whether to compress the directory to a zip file at the end of the
    task run. After compression, it will also delete the directory.
    The default is False.
    """

    @staticmethod
    def setup(directory: str, **kwargs):
        """
        This abstract method is ran before the command is actually executed. This
        allows for some pre-processing, such as writing input files or any other
        analysis.

        When writing a custom S3Task, you can overwrite this method. The only
        criteria is that you...

        1. include `directory` as the 1st input parameter and add `**kwargs`
        2. decorate your method with `@staticmethod` or `@classmethod`

        These criteria allow for compatibility with higher-level functinality.

        You should never call this method directly unless you are debugging. This
        is becuase `setup` is normally called within the `run` method.

        Some tasks don't require a `setup` method, so by default, this method
        does nothing but "pass".

        #### Parameters

        - `directory`:
            The directory to run everything in. Must exist already.

        - `**kwargs`:
            Extra kwargs that may be passed to some function within. Because
            Simmate prefers fixed settings for their workflows, this is typically
            not used, but instead, keywords should be explicitly defined when
            writing a setup method.
        """
        pass

    @classmethod
    def setup_restart(directory: str, **kwargs):
        """
        This method is used instead of `setup` when is_restart=True is passed
        to the run/run_config methods.

        This abstract method is ran before the command is actually executed. This
        allows for some pre-processing, such as writing input files or any other
        analysis.

        When writing a custom S3Task, you can overwrite this method. The only
        criteria is that you...

        1. include `directory` as the 1st input parameter and add `**kwargs`
        2. decorate your method with `@staticmethod` or `@classmethod`

        These criteria allow for compatibility with higher-level functinality.

        You should never call this method directly unless you are debugging. This
        is becuase `setup_restart` is normally called within the `run` method.

        Some tasks don't require a `setup_restart` method, so by default, this
        method does nothing but "pass".

        #### Parameters

        - `directory`:
            The directory to run everything in. Must exist already.

        - `**kwargs`:
            Extra kwargs that may be passed to some function within. Because
            Simmate prefers fixed settings for their workflows, this is typically
            not used, but instead, keywords should be explicitly defined when
            writing a setup method.
        """
        pass

    @classmethod
    def _check_input_files(cls, directory: str, raise_if_missing: bool = True):
        """
        Make sure that there are the proper input files to run this calc
        """

        filenames = [os.path.join(directory, file) for file in cls.required_files]
        if not all(os.path.exists(filename) for filename in filenames):
            if raise_if_missing:
                raise Exception(
                    "Make sure your `setup` method directory source is set up correctly"
                    "The following files must exist in the directory where "
                    f"this task is ran but some are missing: {cls.required_files}"
                )
            return False  # indicates something is missing
        return True  # indicates all files are present

    @classmethod
    def execute(cls, directory: str, command: str) -> List[Tuple[str]]:
        """
        This calls the command within the target directory and handles all error
        handling as well as monitoring of the job.

        You should never call this method directly unless you are debugging. This
        is becuase `execute` is normally called within the `run` method.

        Some tasks don't require a setup() method, so by default, this method
        does nothing but "pass".

        #### Parameters

        - `directory`:
            The directory to run everything in.
        - `command`:
            The command that will be called during execution.

        #### Returns

        - `corrections`
            A list of tuples where each entry is a error identified and the
            correction applied. Ex: [("ExampleError", "ExampleCorrection")]

        """

        # some error_handlers run while the shelltask is running. These are known as
        # Monitors and are labled via the is_monitor attribute. It's good for us
        # to separate these out from other error_handlers.
        cls.monitors = [handler for handler in cls.error_handlers if handler.is_monitor]

        # in case this is a restarted calculation, check if there is a list
        # of corrections in the current directory and load those as the start point
        corrections_filename = os.path.join(directory, "simmate_corrections.csv")
        if os.path.exists(corrections_filename):
            data = pandas.read_csv(corrections_filename)
            corrections = data.values.tolist()
        # Otherwise we start with zero corrections that we slowly add to. This
        # can be thought of as a table with headers of...
        #   ("applied_errorhandler", "correction_applied")
        else:
            corrections = []

        # ------ start of main while loop ------

        # we can try running the shelltask up to max_corrections. Because only one
        # correction is applied per attempt, you can view this as the maximum
        # number of attempts made on the calculation.
        while len(corrections) <= cls.max_corrections:

            # launch the shelltask without waiting for it to complete. Also,
            # make sure to use common shell commands and to set the working
            # directory.
            #
            # Stderr keyword indicates that we should capture the error if one
            # occurs so that we can report it to the user.
            #
            # The preexec_fn keyword allows us to properly terminate jobs that
            # are launched with parallel processes (such as mpirun). This assigns
            # a parent id to it that we use when killing a job (if an error
            # handler calls for us to do so). This isn't possible on Windows though.
            #
            # OPTIMIZE / BUG: preexec_fn adds about 0.02s overhead to the calculation
            # so we may not want to always use it... Instead we could try to only
            # use it when the command includes "mpirun". Though this may introduce
            # a bug if another is another parallel command used besides mpirun.
            # An example of this might be deepmd which automatically submits
            # things in parallel without calling mpirun up-front.
            process = subprocess.Popen(
                command,
                cwd=directory,
                shell=True,
                preexec_fn=None if platform.system() == "Windows"
                # or "mpirun" not in command  # See bug/optimize comment above
                else os.setsid,
                stderr=subprocess.PIPE,
            )

            # Assume the shelltask has no errors and can retry until proven otherwise
            has_error = False
            allow_retry = True

            # If monitor=True, then we want to supervise this shelltask as it
            # runs. If montors=[m1,m2,...], then we have monitors in place to actually
            # perform the monitoring. If both of these cases are true, then we
            # want to go through the error_handlers to check for errors until
            # the shelltask completes.
            if cls.monitor and cls.monitors:

                # ------ start of monitor while loop ------

                # We want to loop until we find an error and keep track of
                # which loops to run the monitor function on because we don't
                # want them to run nonstop. This variable allows us to monitor
                # checks every Nth loop, while we check the shelltask status
                # on all other loops.
                monitor_freq_n = 0
                while not has_error:
                    monitor_freq_n += 1
                    # Sleep the set amount before checking the shelltask status
                    time.sleep(cls.polling_timestep)

                    # check if the shelltasks is complete. poll will return 0
                    # when it's done, in which case we break the loop
                    if process.poll() is not None:
                        break
                    # check whether we should run monitors on this poll loop
                    if monitor_freq_n % cls.monitor_freq == 0:
                        # iterate through each monitor
                        for error_handler in cls.monitors:
                            # check if there's an error with this error_handler
                            # and grab the error if so
                            error = error_handler.check(directory)
                            if error:
                                # determine if the error handler has a
                                # custom termination method. If not, use our
                                # default one from this class.
                                # The "allow_retry" tells us whether we should
                                # end the job even if we still have an error.
                                # For example, our Walltime handler will tell
                                # us to shutdown and not try anymore -- but
                                # it won't raise an error in order to allow
                                # our workup to run.
                                if not error_handler.has_custom_termination:
                                    # If so, we kill the process but don't apply
                                    # the fix quite yet. That step is done below.
                                    allow_retry = cls._terminate_job(
                                        directory=directory,
                                        process=process,
                                        command=command,
                                    )

                                # Otherwise use the custom termination. An
                                # example of this is for codes where you add
                                # a STOP file to get it to finish rather than
                                # just killing the process. We use this feature
                                # in our VASP Walltime handler.
                                else:
                                    allow_retry = error_handler.terminate_job(
                                        directory=directory,
                                        process=process,
                                        command=command,
                                    )
                                # there's no need to look at the other monitors
                                # so break from the for-loop. We also don't
                                # need to monitor the stagedtask anymore since we just
                                # terminated it or signaled for its graceful
                                # end. So update the while-loop condition.
                                has_error = True
                                break

                # ------ end of monitor while loop ------

            # Now just wait for the process to finish. Note we use communicate
            # instead of the .wait() method. This is the recommended method
            # when we have stderr=subprocess.PIPE, which we use above.
            output, errors = process.communicate()

            # check if the return code is non-zero and thus failed.
            # The 'not has_error' is because terminate() will give a nonzero
            # when a monitor is triggered. We don't want to raise that
            # exception here but instead let the monitor handle that
            # error in the code below.
            if process.returncode != 0 and not has_error:
                # convert the error from bytes to a string
                errors = errors.decode("utf-8")
                # and report the error to the user
                raise NonZeroExitError(
                    f"The command ({command}) failed. The error output was...\n {errors}"
                )

            # Check for errors again, because a non-monitor may be higher
            # priority than the monitor triggered above (if there was one).
            # Since the error_handlers are in order of priority, only the first
            # will actually be applied and then we can retry the calc.
            for error_handler in cls.error_handlers:

                # check if there's an error with this error_handler and grab the
                # error if there is one
                error = error_handler.check(directory)
                if error:
                    # record the error in case it wasn't done so above
                    has_error = True
                    # make a copy of the directory contents and
                    # store as an archive within the same directory
                    make_error_archive(directory)
                    # And apply the proper correction if there is one.
                    # Some error_handlers will even raise an error here signaling
                    # that the stagedtask is unrecoverable and a lost cause.
                    correction = error_handler.correct(directory)
                    # record what's been changed
                    corrections.append((error_handler.name, correction))
                    # break from the error_handler for-loop as we only apply the
                    # highest priority fix and nothing else.
                    break

            # write the log of corrections to file if requested. This is written
            # as a CSV file format and done every while-loop cycle because it
            # lets the user monitor the calculation and error handlers applied
            # as it goes. If no corrections were applied, we skip writing the file.
            if cls.save_corrections_to_file and corrections:
                # compile the corrections metadata into a dataframe
                data = pandas.DataFrame(
                    corrections,
                    columns=["error_handler", "correction_applied"],
                )
                # write the dataframe to a csv file
                data.to_csv(corrections_filename, index=False)

            # If there are no errors, we've finished the calculation and can
            # exit the while loop. Alternatively, some "soft" errors (such as
            # Walltimes) signal us to finish even though there's technically
            # a problem -- they do this will allow_retry=False. Otherwise,
            # just leave everything where it's at and restart the
            # while-loop with a new attempt.
            if not has_error or not allow_retry:
                # break the while-loop
                break

        # ------ end of main while loop ------

        # make sure the while loop didn't exit because of the correction limit
        if len(corrections) >= cls.max_corrections:
            raise MaxCorrectionsError(
                "The number of maximum corrections has been exceeded. Note the final "
                "error and its fix are still listed in the corrections file, but it "
                "was never used."
            )
        # now return the corrections for them to stored/used elsewhere
        return corrections

    @staticmethod
    def _terminate_job(directory: str, process: subprocess.Popen, command: str):
        """
        Stopping the command we submitted can be a tricky business if we are running
        scripts in parallel (such as using mpirun). Different computers and OSs
        can require a different 'kill' command so we establish this separate
        method that can be overwritten.

        If you know your command needs a special case to kill all of its spawn
        processes, you can overwrite this method as well.

        Users should never call this directly becuase this is instead applied
        within the execute() method.

        #### Parameters

        - `directory`:
            The base directory the calculation is taking place in.

        - `process`:
            The process object that will be terminated.

        - `command`:
            the command used to launch the process. This is sometimes useful
            when searching for all running processes under this name.

        #### Returns

        - `allow_retry`:
            if the job should be attempted again when there is a possible fix
            from the correct() method. Defaults to True. In the case where
            allow_retry=False and correct() does NOT raise an error, the
            workup methods will still be called - despite the calculation
            technically having an error.
        """

        # The operating system (Windows, OSX, or Linux) will give us the best guess
        # as to where to access the running command. The results are as you'd expect
        # except for Macs, which returns "Darwin".
        #   Linux: Linux
        #   Mac: Darwin
        #   Windows: Windows
        operating_system = platform.system()

        # The normal line to end a popen process is just...
        #   process.terminate()
        # However this struggles to kill all "child" processes if we are using
        # something like mpirun to run things in parallel. Instead, we use the
        # os module to grab the parent id and send that the termination signal,
        # which is also passed on to all child processes.
        # This command also doesn not work on windows, so I need to address this
        # as well.
        if operating_system != "Windows":
            os.killpg(os.getpgid(process.pid), signal.SIGKILL)
        # note: SIGTERM is the normal signal but I use SIGKILL to try to address
        # permission errors.
        else:
            process.terminate()

        # As an example of an alternative approach to killing a job, here is
        # what Custodian (Materials Project) tries when killing a VASP job
        # submitted through mpirun:
        # try:
        #     os.system("killall vasp")
        # except Exception:
        #     pass

        # TODO: make it so terminate scripts can be loaded from the user's config
        # directory, which will make it so they don't have to overwrite all the
        # classes that inherit from this one.

        # By default, termination allows for restarts, so we return a
        # "allow_retry" value of True.
        return True

    @staticmethod
    def workup(directory: str):
        """
        This method is called at the end of a job, *after* error detection.
        This allows post-processing, such as cleanup, analysis of results,
        etc. This should return the result of the entire job, such as a
        the final structure or final energy calculated.

        When writing a custom S3Task, you can overwrite this method. The only
        criteria is that you...

        1. include `directory` as an input parameters
        2. decorate your method with `@staticmethod` or `@classmethod`

        These criteria allow for compatibility with higher-level functinality.

        You should never call this method directly unless you are debugging. This
        is becuase `workup` is normally called within the `run` method.

        Some tasks don't require a `workup` method, so by default, this method
        doesn't nothing but "pass".

        #### Parameters

        - `directory`:
            The directory to run everything in.
        """
        pass

    @classmethod
    def run_config(
        cls,
        directory: str = None,
        command: str = None,
        is_restart: bool = False,
        **kwargs,
    ):
        """
         Runs the entire staged task (setup, execution, workup), which includes
         supervising during execution.

         Call this method once you have your task initialized. For each run you
         can provide a new structure, directory, or command. For example:

         ``` python
         from simmate.calculator.example.tasks import ExampleTask

         my_result = ExampleTask.run(command=my_command)
         ```

         #### Parameters

         - `command`:
             The command that will be called during execution.

         - `directory`:
             The directory to run everything in. This is passed to the ulitities
             function simmate.ulitities.get_directory

        - `is_restart`:
            whether or not this calculation is a continuation of a previous run
            (i.e. a restarted calculation). If so, the `setup_restart` will be
            called instead of the setup method. Extra checks will be made to
            see if the calculation completed already too.

         - `**kwargs`:
             Any extra keywords that should be passed to the setup() method.

        #### Returns

         - a dictionary of the result, corrections, and working directory used
         for this task run
        """

        # because the command is something that is frequently changed at the
        # workflow level, then we want to make it so the user can set it for
        # each unique task.run() call. Otherwise we grab the default from the
        # class attribute
        if not command:
            command = cls.command

        # establish the working directory
        directory = get_directory(directory)

        # When handling restarted calculations, we check for the final summary
        # file and if it exists, we know the calculation has already completed
        # (and therefore doesn't require a restart). This helps handle nested
        # workflows where we don't know which task to restart at.
        summary_filename = os.path.join(directory, "simmate_summary.yaml")
        is_complete = os.path.exists(summary_filename)
        is_dir_setup = cls._check_input_files(directory, raise_if_missing=False)

        # run the setup stage of the task, where there is a unique method
        # if we are picking up from a previously paused run.
        if (not is_restart and not is_complete) or not is_dir_setup:
            cls.setup(directory=directory, **kwargs)
        elif is_restart and not is_complete:
            cls.setup_restart(directory=directory, **kwargs)
        else:
            print("Calculation is already completed. Skipping setup.")

        # now if we have a restart OR have an incomplete calculation that is being
        # restarted, we can check our files and run the external program
        if not is_restart or not is_complete:

            # make sure proper files are present
            cls._check_input_files(directory)

            # run the shelltask and error supervision stages. This method returns
            # a list of any corrections applied during the run.
            corrections = cls.execute(directory, command)
        else:
            print("Calculation is already completed. Skipping execution.")

            # load the corrections from file for reference
            corrections_filename = os.path.join(directory, "simmate_corrections.csv")
            if os.path.exists(corrections_filename):
                data = pandas.read_csv(corrections_filename)
                corrections = data.values.tolist()
            else:
                corrections = []
            # OPTIMIZE: this same code is at the start of the execute method.
            # Consider making it into a utility.

        # run the workup stage of the task. This is where the data/info is pulled
        # out from the calculation and is thus our "result".
        result = cls.workup(directory=directory)

        # if requested, compresses the directory to a zip file and then removes
        # the directory.
        if cls.compress_output:
            make_archive(directory)

        # Grab the prefect flow run id. Note, when not ran within a prefect
        # flow, there won't be an id. We therefore need this try/except
        try:
            prefect_context = get_run_context()
            flow_run_id = str(prefect_context.task_run.flow_run_id)
        except MissingContextError:
            flow_run_id = None

        # Return our final information as a dictionary
        return {
            "result": result,
            "corrections": corrections,
            "directory": directory,
            "prefect_flow_run_id": flow_run_id,
        }

    @classmethod
    @cache
    def to_prefect_task(cls) -> Task:
        """
        Converts this Simmate s3task into a Prefect task
        """

        # Build the Task object directly instead of using prefect's @task decorator
        task = Task(
            fn=cls.run_config,
            name=cls.__name__,
        )

        # as an extra, we set this attribute to the prefect task instance, which
        # allows us to access the source Simmate S3Task easily with Prefect's
        # context managers.
        task.simmate_s3task = cls

        return task

    @classmethod
    def run(cls, **kwargs):
        """
        A convience method to run this task as a registered task in a prefect context.
        """
        task = cls.to_prefect_task()
        state = task.submit(**kwargs)

        # We don't want to block and wait because this might disable parallel
        # features of subflows. We therefore return the state and let the
        # user decide if/when to block.
        # result = state.result()

        return state

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class.

        By default, this will just grab the class's __dict__ attribute but this
        can be overwritten to show only relevent information.
        """
        return dict(cls.__dict__)

    @classmethod
    def print_config(cls):
        """
        Takes the result of get_config and prints it in a yaml format that is
        easier to read.
        """
        config = cls.get_config()
        print(yaml.dump(config))


# Custom errors that indicate exactly what causes the S3Task to exit.


class MaxCorrectionsError(Exception):
    pass


class NonZeroExitError(Exception):
    pass



# -*- coding: utf-8 -*-

import pytest

from simmate.conftest import copy_test_files
from simmate.workflow_engine import S3Task
from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.workflows.diffusion.all import (
    Diffusion__Vasp__NebAllPaths,
)


@pytest.mark.slow
@pytest.mark.prefect_db
@pytest.mark.django_db
def test_neb(sample_structures, tmpdir, mocker):

    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="all_paths",
    )

    # For testing, look at I- diffusion in Y2CF2
    structure = sample_structures["Y2CI2_mp-1206803_primitive"]

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(Potcar, "to_file_from_type", return_value=None)

    # We also don't want to run any commands -- for any task. We skip these
    # by having the base S3task.execute just return an empty list (meaning
    # no corrections were made).
    mocker.patch.object(S3Task, "execute", return_value=[])

    # Don't check for proper input files because POTCARs will be missing
    mocker.patch.object(S3Task, "_check_input_files", return_value=None)

    # run the workflow and make sure it handles data properly.
    state = Diffusion__Vasp__NebAllPaths.run(
        structure=structure,
        migrating_specie="I",
        command="dummycmd1; dummycmd2; dummycmd3",
        directory=str(tmpdir),
    )
    assert state.is_completed()



# -*- coding: utf-8 -*-

"""
To prevent database-connection leaks, we want to setup django upfront when
a worker is started -- rather than have each task connect separately.

You can use this script like so...

``` python 
    from dask.distributed import Client
    client = Client(preload="simmate.configuration.dask.connect_to_database")
```

Note, there is also a high-level utility for this, which can be easier to
remember:
    
``` python
from simmate.utilities import get_dask_client
client = get_dask_client()
"""

# --------------------------------------------------------------------------------------


# First setup django settings for simmate
from simmate.database import connect

# The settings (including the database) are all set up now, but django doesn't
# actually connect to the database until a query is made. So here, we do a
# very simple query that should work for any django database. We don't actaully
# need the output. We just want make a call that establishes a connection.
# Let's just use the ContentType table because it's typically small.
from django.contrib.contenttypes.models import ContentType

# and make a quick query
ContentType.objects.count()


# --------------------------------------------------------------------------------------


# BUG: do I want to make a connection first? It looks like if the connection
# closes at any point, then all following tasks will fail -- that is, once
# the connection closes, no following tasks reopen it. Maybe if I don't establish
# the connection upfront, I can just let every task make their own.
# The error I get is...
#   Unexpected error: InterfaceError('connection already closed') dask
# This has only happened when I submit >5,000 flow runs, so there may be too
# many connections opened at that point. I'm not sure if that's actually the
# case though.

# BUG: for scaling Dask to many workers, I initially ran into issues of "too many
# files open". This is addressed in Dask's FAQ:
#   https://distributed.dask.org/en/latest/faq.html#too-many-open-file-descriptors
# To summarize the fix...
#   (1) check the current soft limit (soft = no sudo permissions) for  files
#       ulimit -Sn
#   (2) to increase the softlimit, edit the limits.conf file and add one line
#       sudo nano /etc/security/limits.conf
#           # add this line below
#           * soft nofile 10240
#   (3) close and reopen the terminal
#   (4) confirm we changed the limit
#       ulimit -Sn
#
# This may also be a leak of sockets being left open by Dask:
#   (1) get the PID of the running process with
#       ps -aef | grep python
#   (2) look at the fd's (file opened) by the given process
#       cd /proc/<PID>/fd; ls -l
#   (3) count the number of files opened by the given process
#       ls /proc/<PID>/fd/ | wc -l
#   (4) view overall stats with
#       cat /proc/<PID>/net/sockstat
#   (5) another option to list open files is
#       lsof -p <PID> | wc -l
#
# Whenever I see a heartbeat fail, I also see a massive jump in the number of
# files opened by the process. I believe zombie prefect runs are creating
# a socket leak.
#

# --------------------------------------------------------------------------------------

# THIS REPRESENTS A PLUGIN FOR A DASK WORKER. I MAY WANT TO SWITCH TO THIS APPROACH
# IN THE FUTURE. See here: https://distributed.dask.org/en/latest/plugins.html

# from dask.distributed import WorkerPlugin


# class DjangoPlugin(WorkerPlugin):
#     def setup(self, worker):

#         print("PRELOADING DJANGO TO DASK WORKER")

#         # First setup django settings for simmate
#         from simmate.database import connect  # ensures setup

#         # The settings (including the database) are all set up now, but django doesn't
#         # actually connect to the database until a query is made. So here, we do a
#         # very simple query that should work for any django database. We don't actaully
#         # need the output. We just want make a call that establishes a connection.
#         # Let's just use the ContentType table because it's typically small.
#         from django.contrib.contenttypes.models import ContentType

#         # and make a quick query
#         ContentType.objects.count()

#     def teardown(self, worker):

#         print("SHUTING DOWN DASK WORKER CONNECTIONS TO DJANGO WORKER")

#         # grab all of the existing database connections for this thread
#         from django.db import connections

#         # close all of these connections to prevent connection leaking
#         connections.close_all()


# REGISTER WITH....
# # def dask_setup(scheduler):
# #     plugin = DjangoPlugin()
# #     client.register_worker_plugin(plugin)  # doctest: +SKIP
# OR...
# from dask.distributed import Client
# client = Client(cluster.scheduler.address)
# from simmate.configuration.dask.connect_to_database import DjangoPlugin
# plugin = DjangoPlugin()
# client.register_worker_plugin(plugin)



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from .incorrect_smearing import IncorrectSmearing
from .mesh_symmetry import MeshSymmetry
from .unconverged import Unconverged
from .nonconverging import NonConverging
from .potim import Potim
from .positive_energy import PositiveEnergy
from .frozen import Frozen
from .large_sigma import LargeSigma

from .tetrahedron_mesh import TetrahedronMesh
from .rotation_matrix import RotationMatrix
from .brmix import Brmix
from .zpotrf import Zpotrf
from .subspace_matrix import SubspaceMatrix
from .incorrect_shift import IncorrectShift
from .real_optlay import RealOptlay
from .tetirr import Tetirr
from .rotation_matrix_nonint import RotationNonIntMatrix
from .long_vector import LongVector
from .triple_product import TripleProduct
from .pricel import Pricel
from .brions import Brions
from .zbrent import Zbrent
from .insufficient_bands import InsufficientBands
from .pssyevx import Pssyevx
from .eddrmm import Eddrmm
from .edddav import Edddav
from .edwav import Edwav
from .zheev import Zheev
from .elf_kpar import ElfKpar
from .rhosyg import Rhosyg
from .posmap import Posmap
from .point_group import PointGroup
from .symprec_noise import SymprecNoise
from .walltime import Walltime



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Zheev(ErrorHandler):

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["ERROR EDDIAG: Call to routine ZHEEV failed!"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # We make a fix only if ALGO is set less to Fast
        if incar.get("ALGO", "Fast") != "Exact":
            incar["ALGO"] = "Exact"
            correction = "switched Algo to Exact"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.website.core_components.filters import Structure
from simmate.database.base_data_types.nudged_elastic_band import (
    DiffusionAnalysis as DiffusionAnalysisTable,
    MigrationHop as MigrationHopTable,
    MigrationImage as MigrationImageTable,
)


class DiffusionAnalysis(Structure):
    class Meta:
        model = DiffusionAnalysisTable
        fields = dict(
            migrating_specie=["exact"],
            vacancy_mode=["exact"],
            atomic_fraction=["range"],
            barrier_cell=["range"],
            npaths_involved=["range"],
            **Structure.get_fields()
        )


class MigrationHop(filters.FilterSet):
    class Meta:
        model = MigrationHopTable
        fields = dict(
            site_start=["exact"],
            site_end=["exact"],
            number=["range"],
            length=["range"],
            dimension_path=["range"],
            dimension_host_lattice=["range"],
            energy_barrier=["range"],
        )


class MigrationImage(Structure):
    class Meta:
        model = MigrationImageTable
        fields = dict(
            number=["exact"],
            force_tangent=["range"],
            energy=["range"],
            structure_distance=["range"],
            **Structure.get_fields()
        )



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.database.base_data_types import Forces as ForcesTable


class Forces(filters.FilterSet):
    class Meta:
        model = ForcesTable
        fields = dict(
            site_force_norm_max=["range"],
            site_forces_norm_per_atom=["range"],
            lattice_stress_norm=["range"],
            lattice_stress_norm_per_atom=["range"],
        )
